""" ARDI Driver Base Class

The 'Base' class encapsulates all of the essential startup and control behaviours of an ARDI LIVE driver.

Each 'Base' contains multiple 'Core' objects, which in turn contain a single source instance.

"""
import sys
import platform
import logging
import logging.handlers
import threading
import driverhttp
import drivercore
import argparse
import pycurl
import xmltodict
import time
import traceback
from StringIO import StringIO
#from twisted.web import server,http,resource
#import twisted

class iologger:
    def __init__(self,logger,level):
        self.logger = logger
        self.linebuf = ''
        self.log_level = level
        
    def write(self,buf):
        for line in buf.rstrip().splitlines():
            self.logger.log(self.log_level,line.rstrip())


def corethread(core, servername, assetid):
    """ Thread body for a single Source.

    This is a basic utility function to launch a single 'source' inside this driver.

    \param core: The core that this driver will use
    \param servername: The name/IP of the system that this driver will report to
    \param assetid: The ARDI asset ID number of the source"""
    core.Launch(servername, assetid)

def webservice(core):
    """ Starts the IPC (socket) service used to communicate with and manage this driver """
    core.logger.info('Starting IPC Service')
    import ipcservice
    
    ipc = ipcservice.DriverIPC()
    ipc.core = core
    ipc.port = core.hostport
    if ipc.port == 0:
        ipc.port = 8081
        
    try:
        ipc.start()
    except KeyboardInterrupt:
        pass    

class ardisourcedetail:
    def __init__(self):
        self.id = 0
        self.subport = 5336
        self.subhost = "localhost"
        self.mode = 0
        self.profile = 1

class ardidriver:
    """ The base class for a driver.

    This class is responsible for all stages of driver life-cycle.

    It also communicates directly with ARDI to get details on what data sources it should serve via HTTP web requests """
    
    def __init__(self):
        ##The port on which this drivers IPC interface will run
        self.hostport = 8081        

        ##Used for future 3rd party integration
        self.base = ''

        ##The shell to be used to execute scripts (mostly required to launch Python in Windows)
        self.shell = ''        

        ##The list of currently running core threads
        self.threads = []

        ##The list of local assets
        self.assets = []
        
    def initlogger(self):
        
        if hasattr(self,'logger'):
            return
        
        """
        Prepare Logging
            
        This function sets up logging (with os-specific paths) for the rest of the driver.
        """
        logging.basicConfig(level=logging.INFO)

        ##The log file that represents the driver process as a whole
        self.logger = logging.getLogger("Driver " + str(self.hostport))
            
        fname = "/var/log/ardi/driver-" + str(self.hostport) + ".log"
        if (platform.system() == 'Windows'):
                fname = "c:\\windows\\temp\\ardi-driver-" + str(self.hostport) + ".log"
            
        loghandler = logging.handlers.WatchedFileHandler(fname)
        loghandler.setLevel(logging.INFO)
            
        # create a logging format
            
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
        loghandler.setFormatter(formatter)
            
        # add the handlers to the logger
            
        self.logger.addHandler(loghandler)

        self.logger.info('----DRIVER RESTART----')
        self.logger.info('Starting Driver')

        redir = iologger(self.logger,logging.ERROR)
        sys.stderr = redir
        
    def start(self,driverfactory, port=None):
        """Start the Driver

        This function is called once all initialisation is complete. It launches sources and the IPC service.

        \param driverfactory: The connector factory class
        \param port: The port that the server is to run off (deprecated)"""
        #self.initlogger()

        ##The driver factory is a class that produces new driver instances        
        self.driverfactory = driverfactory

        self.LoadConfig()
        
        webservice(self)
        
        self.StopDrivers()
        
    def LoadConfig(self):
        """
        Loads driver configuration from the server
        
        This function gathers command-line arguments and then loads configuration details from the web.
        """
        arglist = sys.argv
        if len(arglist) >= 2:
            if arglist[1] == 'start':                
                arglist = arglist[2:]
            else:
                arglist = arglist[1:]

        parser = argparse.ArgumentParser(description='ARDI Driver')
        parser.add_argument('driverid',metavar='PortNo',type=int, help='Unique ID/port for the driver', nargs='?')
        parser.add_argument('server',metavar='Server', help='Server Name', nargs='?', default='localhost')
        parser.add_argument('assetid',metavar='AssetNo',type=int, help='Asset IDs to act as a driver for', nargs='?')
        parser.add_argument('--multiple',dest='multi',action='store_const',const=True,default=False,help='Allows this driver to connect to multiple devices')        
        parser.add_argument('--script','-r',dest='script',default='')
        parser.add_argument('--shell','-l',dest='shell',default='')
        parser.add_argument('--config','-c',dest='configfile',default='')        
        parser.add_argument('--debuglevel',dest='debuglevel',default=2)
        args = parser.parse_args(arglist)
        
        #self.logger.debug(args)
        
        servername = args.server
        self.hostport = int(args.driverid)
        
        self.initlogger()
        
        self.logger.info('Driver ' + str(self.hostport) + ' feeds server ' + servername)

        ##The script that should be run on every successful data fetch
        self.script = args.script
        self.shell = args.shell        

        ##A list of all currently running driver instances
        self.drivers = {}

        ##A dictionary that allows fast association of asset IDs to data points
        self.assetids = []
        if args.assetid != None:
            for assetid in args.assetid:
                self.assetids.append(assetid)
        else:
            response = StringIO()

            sleeptimer = 10
            while (True):
                #try:
                fullurl = "http://" + servername + "/api/driverdata?port=" + str(self.hostport)

                    #print "Fetching " + fullurl

                try:
                    c = pycurl.Curl()
                    c.setopt(c.URL,fullurl)
                    c.setopt(c.WRITEFUNCTION,response.write)
                    c.setopt(c.TIMEOUT,5)
                    c.setopt(c.CONNECTTIMEOUT,5)
                    c.perform()
                    c.close()
                    break
                except:
                    sleeptimer += 10
                    if (sleeptimer > 60):
                        sleeptimer = 60

                    self.logger.info("ARDI Not Responding. Trying again in " + str(sleeptimer) + " seconds")

                    time.sleep(sleeptimer)                    
                finally:
                    c.close()
        
            self.parseconfig(response.getvalue())
            
            self.logger.info("Load Default Configuration From Server")
        
        #for assetid in self.assetids:
        for asset in self.assets:
            self.logger.info("Creating driver instance for " + str(asset.id))
            instance = self.driverfactory.createinstance()
            core = drivercore.ardicore(instance)
            core.base = self
            core.source = args.configfile

            subtarget = asset.subhost
            if (asset.subhost == "localhost"):
                #Strip out junk and copy the IP of the original host...
                
                subtarget = servername
                ps = subtarget.find('/');
                if (ps != -1):
                    subtarget = subtarget[0:ps]

                ps = subtarget.find(':');
                if (ps != -1):
                    subtarget = subtarget[0:ps]
                    
            asset.subhost = subtarget
                    
            core.info = asset

            self.logger.info("Source reports to " + str(asset.subhost) + ":" + str(asset.subport) + " & " + str(servername));
            
            self.startcorethread(core, servername, asset.id)
            self.drivers[str(asset.id)] = core

        #print self.drivers
        #webservice(self)
        
    def parseconfig(self,xml):
        """Imports the configuration XML file

        This function loads the list of data points within the XML text provided. The function is usually called from LoadConfig

        \param xml: The XML text to process"""
        print xml
        result = xmltodict.parse(xml)

        try:
            #Core Parameters
            if '@id' in result['driverdata']['source']:
                self.assetids.append(result['driverdata']['source']['@id'])
                
                ast = ardisourcedetail()                
                ast.id = result['driverdata']['source']['@id']
                ast.subport = result['driverdata']['source']['@subport']
                ast.subhost = result['driverdata']['source']['@host']
                ast.mode = result['driverdata']['source']['@mode']
                ast.profile = result['driverdata']['source']['@profile']
                
                self.assets.append(ast)
            else:
                for itm in result['driverdata']['source']:                    
                    self.assetids.append(itm['@id'])
                    
                    ast = ardisourcedetail()                
                    ast.id = itm['@id']
                    ast.subport = itm['@subport']
                    ast.subhost = itm['@host']
                    ast.mode = itm['@mode']
                    ast.profile = itm['@profile']   
                    self.assets.append(ast)
        except:
            #print "Exception Processing"
            #traceback.print_exc()
            pass
    
    def StopDrivers(self):
        """
        Stops all of the Cores under this base driver

        This function is primarily used during driver shutdown or restart
        """
        for drv in self.drivers:
            self.drivers[drv].Stop()
        
        self.assets = []
        self.drivers = []
        self.threads = []
        self.logger.info('All Drivers Stopped')
        pass
        
    def startcorethread(self,core, servername, assetid):
        """Starts a single Core thread
        
        \param core: The core to run in this thread
        \param servername: The name of the server to report all data to
        \param assetid: The asset ID number of the source
        """
        t = threading.Thread(target=corethread, args = (core, servername, assetid))
        t.daemon = True
        t.start()
        #self.threads[int(assetid)] = t
    
    def Pause(self, assetid):
        """Pauses a given data source

        This function is called from the drivers IPC service

        \param assetid: The ID of the source that you wish to reload"""
        self.logger.info('Pause Requested via IPC')
        self.drivers[assetid].Pause()
    
    def Reload(self, assetid):
        """Reload a single source

        This function is called from the IPC service, usually due to a change in the underlying ARDI
        data bindings - such as the addition of a new address, or change of an existing one.

        \param assetid: The ID of the source that you wish to reload"""
        if assetid in self.drivers.keys():
            self.logger.info('Source Reload Requested via IPC On Source ' + str(assetid))
            self.drivers[assetid].Reload()
        else:
            self.logger.error('Source Reload Requested on Invalid Source ( ' + str(assetid) + " )")

    def Restart(self):
        """Forces all sources to reload and restart

        This function is called from the drivers IPC service"""
        self.logger.info('Driver Restart Requested via IPC')
        self.StopDrivers()
        self.assets = []
        self.LoadConfig()
        pass
        
    def Spinup(self, assetid):
        """Requests that a new source be started

        This function is called from the drivers IPC service

        \param assetid: The ID of the service you wish to initialise"""
        self.logger.info('Spinup Request - Performing Restart')
        self.Restart()
        
        return
    
        if assetid in self.drivers.keys():
            self.logger.info('Spinup Request on Existing Source ' + str(assetid))
            self.drivers[assetid].Reload()
        else:        
            self.logger.info("Spinup Request on New Source " + str(assetid))
            instance = self.driverfactory.createinstance()
            core = drivercore.ardicore(instance)
            core.base = self
            core.source = args.configfile
            self.startcorethread(core, servername, assetid)
            self.drivers[str(assetid)] = core
    
    def Resume(self,assetid):
        """Resume a paused service

        This function is called from the drivers IPC service

        \param assetid: The ID of the service you wish to pause """
        self.drivers[assetid].Resume()

    def DebugMode(self,assetid,action):
        """Enable or disable debugging on a particular source.

        This function is called from the drivers IPC service

        \param assetid: The ID of the service you wish to debug
        \param action: '1' to enable debugging, '0' to disable it"""

        self.logger.info('Driver Debug Mode Request')
        enable = True
        try:
            if (int(action) == 0):
                enable = False
        except:
            pass
        
        log = logging.getLogger('Source '+ str(assetid))
        if len(log.handlers) > 0:
            if enable == True:
                self.logger.info("Turning On Source Logging For Asset " + str(assetid))
                log.handlers[0].setLevel(logging.DEBUG)
                log.setLevel(logging.DEBUG)
            else:
                self.logger.info("Turning Off Source Logging For Asset " + str(assetid))
                log.handlers[0].setLevel(logging.INFO)
                log.setLevel(logging.INFO)
        #else:
        #    print "No Handlers - Log Level Not Changed"
            
        try:
            #print dir(self.drivers[assetid].driver)
            if (hasattr(self.drivers[assetid].driver,'DebugMode')):
                self.logger.info("Running Source-Specific Debugging Routine")                
                self.drivers[assetid].driver.DebugMode(enable)
            else:
                self.logger.info("This Source Does Not Have A Debugging Routine")              
        except Exception as e:
            self.logger.info("Unable To Set DebugMode On Source: " + str(e))
            pass

    def SetUpdateRate(self,assetid,rate):
        """Alters the update rate for the data source

        This function is called from the web interface

        \param assetid: The ID of the service you wish to change
        \param rate: The rate at which to update the data source (in seconds) """
        try:
            self.drivers[assetid].SetUpdateRate(float(rate))
        except:
            pass