import sys

from ardi.driver import drivercore,driverbase

from pymodbus.client.sync import ModbusTcpClient as ModbusClient
try:    
    from pymodbus.payload import PayloadDecoder as BinaryPayloadDecoder
except:
    from pymodbus.payload import BinaryPayloadDecoder

class ardiliveconnectordemo:
    """An example live connector

    In this case, the connector is for a device that uses the Modbus TCP protocol. Note that any member variables of
    this class are actually private - the system only requires only the connected and polling properties.

    Live connectors MUST implement the SetAddress, Connect, Disconnect, Optimise and Poll (Pull Drver) or Update (Push Driver) functions. """
    def __init__(self):
        """Initialises the connector """
        ##A reference to the core, which contains data and functions that the connector requires.
        self.core = None
        ##True if the conenctor is connected to its data source
        self.connected = False
        ##True (default) if this device is a pull/polling driver
        self.polling = True
    
    def SetAddress(self, addr):
        """Set the address of the connection target

        This is a string that contains all of the details required to connect to the MySQL database.

        \param addr: A colon (:) delimited list of connection properties """
        self.connected = False
        bits = addr.split(":")
        ##The IP address or host name of the device
        self.ip = bits[0]

        ##The port number for the device (0=502)
        self.port = int(bits[1])

        ##The unit number for the device
        self.unit = int(bits[2])
    
    def Connect(self):
        """Connect to the datasource

        \return Returns false on a connection error, True on success """
        self.connected = False

        ##The Modbus TCP client session
        self.client = ModbusClient(self.ip, port=self.port)
        try:
            success = self.client.connect()
            if success == False:
                return False
            self.connected = True
            return True
        except:
           return False
    
    def Disconnect(self):
        """Disconnect from the datasource """
        self.connected = False
        self.client.close()
        
    def Poll(self):
        """Poll for new data

        This function is called to make the connector fetch new data from the source.

        The connector queries the items in 'self.core.pointlist' and if changes are detected it calls
        the 'self.core.UpdateData' function to pass this new value into the drivers buffer, in preperation
        to be transmitted to ARDI (the actual data updates are buffered so that there is only one ARDI update per
        driver poll, rather than dozens or hundreds)"""
        for item in self.core.pointlist:
            bits = item.address.split('|')
            mode = 0
            address = 0
            
            if len(bits) == 1:
                address = int(bits[0])
            else:
                address = int(bits[0])
                mode = int(bits[1])
                
            if address < 10000:
                rr = self.client.read_coils(address,1)
                self.core.UpdateData(item.code,rr.bits[0])
                continue
            
            if address < 20000:
                rr = self.client.read_discrete_inputs(address-10000,1)
                self.core.UpdateData(item.code,rr.bits[0])
                continue
            
            ln = 1
            if (mode == 1):
                ln = 2
            elif (mode == 2):
                ln = 4
            elif (mode == 3):
                ln = int(bits[2])
            
            if address < 40000:
                rr = self.client.read_input_registers(address-30000,ln)
            else:				
				rr = self.client.read_holding_registers(address-40000,ln)
            
            if mode == 0:
                print 'Reading ' + str(address)				
                print str(rr.registers[0])
                self.core.UpdateData(item.code,rr.registers[0])
                return
            
            en = '<';
            decoder = BinaryPayloadDecoder.fromRegisters(rr.registers,endian=en)
            
            if mode == 1:
                val = decoder.decode_32bit_int()				
                self.core.UpdateData(item.code,val)
            if mode == 10:
                val = decoder.decode_32bit_float()
                self.core.UpdateData(item.code,val)
                return
            if mode == 11:
                val = decoder.decode_64bit_float()
                self.core.UpdateData(item.code,val)
            if mode == 20:
                val = decoder.decode_string(int(bits[2]))
                self.core.UpdateData(item.code,val)
                
    def Optimise(self):
        """Optimise and update the query

        This function is called when the data that is being scanned from the data source changes, either
        due to user interaction (such as adding or removing data links) or due to changes in interest when
        using interest-based data loading.

        Use this opportunity to prepare query strings, subscribe to notifications or optimise your
        fetch requests to use the minimum network bandwidth.

        TODO: This driver currently sends one request per point, which is highly inefficient. Later updates
        will apply automatic optimisation here to reduce the number of Modbus TCP queries."""
        pass
        
class ardiliveconnectorfactorydemo:
    """Demo live connector factory

    Every driver (live or historian) requires a 'Connector Factory' - a class that produces new instances of the
    class that manages a connection to a data source.

    It must implement the 'createinstance' function. """
    def createinstance(self):
        """Create a single connector instance """
        return ardiliveconnectordemo()
            
if __name__ == "__main__":
    sdf = ardiliveconnectorfactorydemo()
    base = driverbase.ardidriver()
    base.start(sdf)